from enum import IntEnum

from pymchelper.writers.excel import ExcelWriter
from pymchelper.writers.inspector import Inspector
from pymchelper.writers.plots import GnuplotDataWriter, PlotDataWriter, ImageWriter
from pymchelper.writers.shieldhit import TxtWriter
from pymchelper.writers.sparse import SparseWriter
from pymchelper.writers.trip98 import TripCubeWriter, TripDddWriter
from pymchelper.writers.hdf import HdfWriter


class Converters(IntEnum):
    """
    Available converters
    """
    txt = 0
    plotdata = 1
    gnuplot = 2
    image = 3
    tripcube = 4
    tripddd = 5
    excel = 6
    sparse = 7
    inspect = 8
    hdf = 9

    @classmethod
    def _converter_mapping(cls, item):
        return {
            cls.txt: TxtWriter,
            cls.gnuplot: GnuplotDataWriter,
            cls.plotdata: PlotDataWriter,
            cls.image: ImageWriter,
            cls.tripcube: TripCubeWriter,
            cls.tripddd: TripDddWriter,
            cls.excel: ExcelWriter,
            cls.sparse: SparseWriter,
            cls.inspect: Inspector,
            cls.hdf: HdfWriter
        }.get(item)

    @classmethod
    def fromname(cls, name):
        return cls._converter_mapping(Converters[name])

    @classmethod
    def fromnumber(cls, number):
        return cls._converter_mapping[Converters(number)]



from collections import namedtuple
from enum import IntEnum
import logging

import numpy as np
# try to set legacy printing options if working with numpy 1.14 or newer
# on older numpy versions this shouldn't have effect
try:
    np.set_printoptions(legacy="1.13")
except TypeError as e:  # noqa: F841
    pass

logger = logging.getLogger(__name__)


class MeshAxis(namedtuple('MeshAxis', 'n min_val max_val name unit binning')):
    """
    Scoring mesh axis data.

    It can represent an axis variety of scorers:
    x,y or z in cartesian scoring, r, rho or z in cylindrical.
    An axis represents a sequence of ``n`` numbers, defining linear or logarithmic binning.
    This sequence of numbers is not stored in the memory, but can be generated using data property method.
    ``min_val`` is lowest bin left edge, max_val is highest bin right edge
    ``name`` can be used to define physical quantity (i.e. position, energy, angle).
    ``unit`` gives physical units (i.e. cm, MeV, mrad).

    ``MeshAxis`` is constructed as immutable data structure, thus it is possible
    to set field values only upon object creation. Later they are available for read only.

    >>> x = MeshAxis(n=10, min_val=0.0, max_val=30.0, name="Position", unit="cm", binning=MeshAxis.BinningType.linear)
    >>> x.n, x.min_val, x.max_val
    (10, 0.0, 30.0)
    >>> x.n = 5
    Traceback (most recent call last):
    ...
    AttributeError: can't set attribute

    ``binning`` field (use internal ``BinningType.linear`` or ``BinningType.logarithmic``) can distinguish
    log from linear binning
    """
    class BinningType(IntEnum):
        """
        type of axis generator
        """
        linear = 0
        logarithmic = 1

    @property
    def data(self):
        """
        Generates linear or logarithmic sequence of ``n`` numbers.

        These numbers are middle points of the bins
        defined by ``n``, ``min_val`` and ``max_val`` parameters.

        >>> x = MeshAxis(n=10, min_val=0.0, max_val=10.0, name="X", unit="cm", binning=MeshAxis.BinningType.linear)
        >>> x.data
        array([ 0.5,  1.5,  2.5,  3.5,  4.5,  5.5,  6.5,  7.5,  8.5,  9.5])

        Binning may also consist of one bin:
        >>> x = MeshAxis(n=1, min_val=0.0, max_val=5.0, name="X", unit="cm", binning=MeshAxis.BinningType.linear)
        >>> x.data
        array([ 2.5])

        Logarithmic binning works as well, middle bin points are calculated as geometrical mean.
        Here we define 3 bins: [1,4], [4,16], [16,64].
        >>> x = MeshAxis(n=3, min_val=1.0, max_val=64.0, name="X", unit="cm", binning=MeshAxis.BinningType.logarithmic)
        >>> x.data
        array([  2.,   8.,  32.])

        For the same settings as below linear scale gives as expected different sequence:
        >>> x = MeshAxis(n=3, min_val=1.0, max_val=64.0, name="X", unit="cm", binning=MeshAxis.BinningType.linear)
        >>> x.data
        array([ 11.5,  32.5,  53.5])

        For logarithmic axis min_val has to be positive:
        >>> x = MeshAxis(n=3, min_val=-2.0, max_val=64.0, name="X", unit="cm", binning=MeshAxis.BinningType.logarithmic)
        >>> x.data
        Traceback (most recent call last):
        ...
        Exception: Left edge of first bin (-2) is not positive

        :return:
        """
        if self.max_val < self.min_val:
            raise Exception("Right edge of last bin ({:g}) is smaller than left edge of first bin ({:g})".format(
                self.max_val, self.min_val
            ))
        if self.binning == self.BinningType.linear:
            bin_width = (self.max_val - self.min_val) / self.n
            first_bin_mid = self.min_val + bin_width / 2.0  # arithmetic mean
            last_bin_mid = self.max_val - bin_width / 2.0  # arithmetic mean
            return np.linspace(start=first_bin_mid, stop=last_bin_mid, num=self.n)
        elif self.binning == self.BinningType.logarithmic:
            if self.min_val < 0.0:
                raise Exception("Left edge of first bin ({:g}) is not positive".format(self.min_val))
            q = (self.max_val / self.min_val)**(1.0 / self.n)  # an = a0 q^n
            first_bin_mid = self.min_val * q**0.5  # geometrical mean sqrt(a0 a1) = sqrt( a0 a0 q) = a0 sqrt(q)
            last_bin_mid = self.max_val / q**0.5  # geometrical mean sqrt(an a(n-1)) = sqrt( an an/q) = an / sqrt(q)
            try:
                result = np.geomspace(start=first_bin_mid, stop=last_bin_mid, num=self.n)
            except AttributeError:
                # Python3.2 require numpy older than 1.2, such versions of numpy doesn't have geomspace function
                # in such case we calculate geometrical binning manually
                result = np.exp(np.linspace(start=np.log(first_bin_mid),
                                            stop=np.log(last_bin_mid),
                                            num=self.n))
            return result
        else:
            return None


class ErrorEstimate(IntEnum):
    none = 0
    stderr = 1
    stddev = 2


class Detector:
    """
    Detector data including scoring mesh description.

    This class handles in universal way data generated with MC code.
    It includes data (``data`` and ``data_raw`` fields) and optional errors (``error`` and ``error_raw``).
    Detector holds also up to 3 binning axis (``x``, ``y`` and ``z`` fields).
    Scored quantity can be assigned a ``name`` (i.e. dose) and ``unit`` (i.e. Gy).
    Several other fields are also used:
      - nstat: number of simulated histories
      - counter: number of files read to construct detector object
      - corename: common core part of input files defining a name of detector
      - error_type: none, stderr or stddev - error type

    Detector data can be either read from the file (see ``fromfile`` method in ``io`` module
    or constructed directly:

    >>> d = Detector()
    >>> d.x = MeshAxis(n=2, min_val=0.0, max_val=10.0, name="X", unit="cm", binning=MeshAxis.BinningType.linear)
    >>> d.x.data
    array([ 2.5,  7.5])
    >>> d.y = MeshAxis(n=3, min_val=0.0, max_val=150.0, name="Y", unit="cm", binning=MeshAxis.BinningType.linear)
    >>> d.y.data
    array([  25.,   75.,  125.])
    >>> d.z = MeshAxis(n=1, min_val=0.0, max_val=1.0, name="Z", unit="cm", binning=MeshAxis.BinningType.linear)
    >>> d.z.data
    array([ 0.5])
    >>> d.data_raw = np.arange(6)
    >>> d.data.shape
    (2, 3, 1)
    >>> d.data
    array([[[0],
            [1],
            [2]],
           [[3],
            [4],
            [5]]])

    """

    def __init__(self):
        """
        Create dummy detector object.

        >>> d = Detector()
        >>> d.x.data
        array([ nan])
        >>> d.y.data
        array([ nan])
        >>> d.z.data
        array([ nan])
        >>> d.data.shape
        (1, 1, 1)
        >>> d.data
        array([[[ nan]]])
        """
        self.x = MeshAxis(n=1,
                          min_val=float("NaN"),
                          max_val=float("NaN"),
                          name="",
                          unit="",
                          binning=MeshAxis.BinningType.linear)
        self.y = self.x
        self.z = self.x
        self.data_raw = np.array([float("NaN")])
        self.error_raw = np.array([float("NaN")])
        self.name = ""
        self.unit = ""
        self.nstat = 0  # number of histories simulated
        self.counter = 0  # number of files read
        self.corename = ""  # common core for paths of contributing files
        self.error_type = ErrorEstimate.none

    def axis(self, id):
        """
        Mesh axis selector method based on integer id's.

        Instead of getting mesh axis data by calling `d.x`, `d.y` or `d.z` (assuming `d` an object of `Detector`
        class) we can get that data by calling `d.axis(0)`, `d.axis(1)` or `d.axis(2)`. See for example:
        >>> d = Detector()
        >>> d.x = MeshAxis(n=2, min_val=0.0, max_val=10.0, name="X", unit="cm", binning=MeshAxis.BinningType.linear)
        >>> d.y = MeshAxis(n=3, min_val=0.0, max_val=150.0, name="Y", unit="cm", binning=MeshAxis.BinningType.linear)
        >>> d.z = MeshAxis(n=1, min_val=0.0, max_val=1.0, name="Z", unit="cm", binning=MeshAxis.BinningType.linear)
        >>> d.axis(1)
        MeshAxis(n=3, min_val=0.0, max_val=150.0, name='Y', unit='cm', binning=<BinningType.linear: 0>)

        :param id: axis id (0, 1 or 2)
        :return: MeshAxis object
        """
        if id == 0:
            return self.x
        elif id == 1:
            return self.y
        elif id == 2:
            return self.z
        return None

    def plot_axis(self, id):
        """
        Calculate new order of detector axis, axis with data (n>1) comes first
        Axes with constant value goes last.

        Let's take a detector d with YZ scoring.
        >>> d = Detector()
        >>> d.x = MeshAxis(n=1, min_val=0.0, max_val=1.0, name="X", unit="cm", binning=MeshAxis.BinningType.linear)
        >>> d.y = MeshAxis(n=3, min_val=0.0, max_val=150.0, name="Y", unit="cm", binning=MeshAxis.BinningType.linear)
        >>> d.z = MeshAxis(n=2, min_val=0.0, max_val=2.0, name="Z", unit="cm", binning=MeshAxis.BinningType.linear)

        First axis for plotting will be Y (as X axis holds only one bin):
        >>> d.plot_axis(0)
        MeshAxis(n=3, min_val=0.0, max_val=150.0, name='Y', unit='cm', binning=<BinningType.linear: 0>)

        Second axis for plotting will be Z (its the next after Y with n > 1 bins)
        >>> d.plot_axis(1)
        MeshAxis(n=2, min_val=0.0, max_val=2.0, name='Z', unit='cm', binning=<BinningType.linear: 0>)

        Finally the third axis will be X, but it cannot be used for plotting as it has only one bin.
        >>> d.plot_axis(2)
        MeshAxis(n=1, min_val=0.0, max_val=1.0, name='X', unit='cm', binning=<BinningType.linear: 0>)


        :param id: axis number (0, 1 or 2)
        :return: axis object
        """
        plotting_order = (0, 1, 2)
        if self.dimension == 1:
            if self.x.n > 1:
                plotting_order = (0, 1, 2)  # X variable; Y,Z constant
            elif self.y.n > 1:
                plotting_order = (1, 0, 2)  # Y variable; X,Z constant
            elif self.z.n > 1:
                plotting_order = (2, 0, 1)  # Z variable; X,Y constant
        elif self.dimension == 2:
            if self.x.n == 1:
                plotting_order = (1, 2, 0)  # Y,Z variable; X constant
            elif self.y.n == 1:
                plotting_order = (0, 2, 1)  # X,Z variable; Y constant
            elif self.z.n == 1:
                plotting_order = (0, 1, 2)  # X,Y variable; Z constant

            # when SH12A differential scorer is used, we assume that differential
            # quantity should go last
            # a special case is when when X-constant, Y-differential, Z-scored
            # we need to swap X and Z to guarantee that the differential quantity will be last
            if hasattr(self, 'dif_axis') and self.dif_axis == 1:
                plotting_order = (2, 1, 0)

        return self.axis(plotting_order[id])

    @property
    def dimension(self):
        """
        Let's take again detector d with YZ scoring.
        >>> d = Detector()
        >>> d.x = MeshAxis(n=1, min_val=0.0, max_val=1.0, name="X", unit="cm", binning=MeshAxis.BinningType.linear)
        >>> d.y = MeshAxis(n=3, min_val=0.0, max_val=150.0, name="Y", unit="cm", binning=MeshAxis.BinningType.linear)
        >>> d.z = MeshAxis(n=2, min_val=0.0, max_val=2.0, name="Z", unit="cm", binning=MeshAxis.BinningType.linear)
        >>> d.dimension
        2

        :return: number of axes which have more than one point
        """
        return 3 - (self.x.n, self.y.n, self.z.n).count(1)

    @property
    def data(self):
        """
        3-D view of detector data.

        Detector data are stored originally in `data_raw` 1-D array.
        This property provides efficient view of detector data, suitable for numpy-like indexing.

        >>> d = Detector()
        >>> d.x = MeshAxis(n=2, min_val=0.0, max_val=10.0, name="X", unit="cm", binning=MeshAxis.BinningType.linear)
        >>> d.y = MeshAxis(n=3, min_val=0.0, max_val=150.0, name="Y", unit="cm", binning=MeshAxis.BinningType.linear)
        >>> d.z = MeshAxis(n=1, min_val=0.0, max_val=1.0, name="Z", unit="cm", binning=MeshAxis.BinningType.linear)
        >>> d.data_raw = np.arange(6)
        >>> d.data.shape
        (2, 3, 1)
        >>> d.data[1, 2, 0]
        5

        :return: reshaped view of ``data_raw``
        """
        return self.data_raw.reshape((self.x.n, self.y.n, self.z.n))

    @property
    def error(self):
        """
        3-D view of detector error

        For more details see ``data`` property.
        :return:
        """
        return self.error_raw.reshape((self.x.n, self.y.n, self.z.n))


def average_with_nan(detector_list, error_estimate=ErrorEstimate.stderr):
    """
    Calculate average detector object, excluding malformed data (NaN) from averaging.
    :param detector_list:
    :param error_estimate:
    :return:
    """
    # TODO add compatibility check
    result = Detector()
    result.counter = len(detector_list)
    result.data_raw = np.nanmean([det.data_raw for det in detector_list], axis=0)
    if result.counter > 1 and error_estimate != ErrorEstimate.none:
        # s = stddev = sqrt(1/(n-1)sum(x-<x>)**2)
        # s : corrected sample standard deviation
        result.error_raw = np.nanstd([det.data_raw for det in detector_list], axis=0, ddof=1)

        # if user requested standard error then we calculate it as:
        # S = stderr = stddev / sqrt(n), or in other words,
        # S = s/sqrt(N) where S is the corrected standard deviation of the mean.
        if error_estimate == ErrorEstimate.stderr:
            result.error_raw /= np.sqrt(result.counter)  # np.sqrt() always returns np.float64
    else:
        result.error_raw = np.zeros_like(result.data_raw)
    return result



import logging
import numpy as np

logger = logging.getLogger(__name__)


class HdfWriter:
    """
    Supports writing HDF file format.
    HDF is designed to store large amounts of data organized in convenient way.
    One HDF file can handle many single- or multi-dimensional tables.

    """
    def __init__(self, filename, options):
        self.filename = filename
        if not self.filename.endswith(".h5"):
            self.filename += ".h5"

    def write(self, detector):

        try:
            import h5py
        except ImportError as e:
            logger.error("Generating HDF5 files not available on your platform (please install h5py).")
            raise e

        with h5py.File(self.filename, 'w') as f:

            # change units for LET from MeV/cm to keV/um if necessary
            # a copy of data table is made here
            from pymchelper.shieldhit.detector.detector_type import SHDetType
            if detector.dettyp in (SHDetType.dlet, SHDetType.dletg, SHDetType.tlet, SHDetType.tletg):
                data = detector.data * np.float64(0.1)  # 1 MeV / cm = 0.1 keV / um
                if not np.all(np.isnan(detector.error_raw)) and np.any(detector.error_raw):
                    error = detector.error * np.float64(0.1)  # 1 MeV / cm = 0.1 keV / um
            else:
                data = detector.data
                error = detector.error

            # save data
            dset = f.create_dataset("data", data=data, compression="gzip", compression_opts=9)

            # save error (if present)
            if not np.all(np.isnan(detector.error_raw)) and np.any(detector.error_raw):
                f.create_dataset("error", data=error, compression="gzip", compression_opts=9)

            # save metadata
            dset.attrs['name'] = detector.name
            dset.attrs['unit'] = detector.unit
            dset.attrs['nstat'] = detector.nstat
            dset.attrs['counter'] = detector.counter
            dset.attrs['counter'] = detector.counter
            dset.attrs['xaxis_n'] = detector.x.n
            dset.attrs['xaxis_min'] = detector.x.min_val
            dset.attrs['xaxis_max'] = detector.x.max_val
            dset.attrs['xaxis_name'] = detector.x.name
            dset.attrs['xaxis_unit'] = detector.x.unit
            dset.attrs['yaxis_n'] = detector.y.n
            dset.attrs['yaxis_min'] = detector.y.min_val
            dset.attrs['yaxis_max'] = detector.y.max_val
            dset.attrs['yaxis_name'] = detector.y.name
            dset.attrs['yaxis_unit'] = detector.y.unit
            dset.attrs['zaxis_n'] = detector.z.n
            dset.attrs['zaxis_min'] = detector.z.min_val
            dset.attrs['zaxis_max'] = detector.z.max_val
            dset.attrs['zaxis_name'] = detector.z.name
            dset.attrs['zaxis_unit'] = detector.z.unit

        return 0



#!/bin/env python
# $Id: fortran.py 2947 2014-01-17 16:24:41Z bnv $
#
# Copyright and User License
# ~~~~~~~~~~~~~~~~~~~~~~~~~~
# Copyright Vasilis.Vlachoudis@cern.ch for the
# European Organization for Nuclear Research (CERN)
#
# All rights not expressly granted under this license are reserved.
#
# Installation, use, reproduction, display of the
# software ("flair"), in source and binary forms, are
# permitted free of charge on a non-exclusive basis for
# internal scientific, non-commercial and non-weapon-related
# use by non-profit organizations only.
#
# For commercial use of the software, please contact the main
# author Vasilis.Vlachoudis@cern.ch for further information.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the
#    distribution.
#
# DISCLAIMER
# ~~~~~~~~~~
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
# NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY, OF
# SATISFACTORY QUALITY, AND FITNESS FOR A PARTICULAR PURPOSE
# OR USE ARE DISCLAIMED. THE COPYRIGHT HOLDERS AND THE
# AUTHORS MAKE NO REPRESENTATION THAT THE SOFTWARE AND
# MODIFICATIONS THEREOF, WILL NOT INFRINGE ANY PATENT,
# COPYRIGHT, TRADE SECRET OR OTHER PROPRIETARY RIGHT.
#
# LIMITATION OF LIABILITY
# ~~~~~~~~~~~~~~~~~~~~~~~
# THE COPYRIGHT HOLDERS AND THE AUTHORS SHALL HAVE NO
# LIABILITY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL,
# CONSEQUENTIAL, EXEMPLARY, OR PUNITIVE DAMAGES OF ANY
# CHARACTER INCLUDING, WITHOUT LIMITATION, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES, LOSS OF USE, DATA OR PROFITS,
# OR BUSINESS INTERRUPTION, HOWEVER CAUSED AND ON ANY THEORY
# OF CONTRACT, WARRANTY, TORT (INCLUDING NEGLIGENCE), PRODUCT
# LIABILITY OR OTHERWISE, ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGES.
#
# Author:	Vasilis.Vlachoudis@cern.ch
# Date:	24-Oct-2006

import struct

__author__ = "Vasilis Vlachoudis"
__email__ = "Vasilis.Vlachoudis@cern.ch"


def skip(f):
    """
    Skip a fortran block from a binary file
    :param f: file to read from
    :return: size, None for EOF
    """
    blen = f.read(4)
    if len(blen) == 0:
        return 0
    (size, ) = struct.unpack("=i", blen)
    f.seek(size, 1)
    blen2 = f.read(4)
    if blen != blen2:
        raise IOError("Skipping fortran block")
    return size


def read(f):
    """
    Read a fortran structure from a binary file
    :param f: file to read from
    :return: data, None for EOF
    """
    blen = f.read(4)
    if len(blen) == 0:
        return None
    (size, ) = struct.unpack("=i", blen)
    data = f.read(size)
    blen2 = f.read(4)
    if blen != blen2:
        if not f.closed:
            f.close()
        raise IOError("Reading fortran block")
    return data


def write(f, d):
    """
    Write a block of data (string) to file as a fortran block
    :param f: file to write to
    :param d: data to write
    :return: output of last write statement
    """
    f.write(struct.pack("=i", len(d)))
    f.write(d)
    return f.write(struct.pack("=i", len(d)))



import sys
from itertools import product

from pymchelper.flair import Input
from pymchelper.flair.Input import Card
from pymchelper.shieldhit.detector.geometry import CarthesianMesh


def generate_fluka_file(output_filename):
    """
    Compose programatically fl_sim.inp Fluka input file
    with many combinations of detector and mesh types.

    Fluka input file is saved in the default format (as Flair saves it), namely:
      - fixed format is used everywhere, except:
      - free format for geometry, forced by COMBNAME paramater in GEOBEGIN card
      - names instead of numbers are used as identifies of particles, materials and regions

    :param output_filename: output filename
    :return: generated Fluka object
    """

    # create empty input file object
    input_configuration = Input.Input()

    # add title
    input_configuration.addCard(Card("TITLE", extra="proton beam simulation"))

    # add default physics settings
    input_configuration.addCard(Card("DEFAULTS", what=["HADROTHE"], comment="default physics for hadron therapy"))

    # add beam characteristics, see http://www.fluka.org/fluka.php?id=man_onl&sub=12
    beam = Card("BEAM", comment="beam characteristics")
    beam.setSdum("PROTON")  # SDUM    = beam particle name.
    beam.setWhat(1, -0.06)  # WHAT(1) < 0.0 : average beam kinetic energy in GeV
    beam.setWhat(4, 2.0)  # WHAT(4) > 0.0 : If WHAT(6) > 0.0, beam width in x-direction in cm,
    # The beam profile is assumed to be rectangular
    beam.setWhat(5, 2.0)  # WHAT(5) > 0.0 : If WHAT(6) > 0.0, beam width in y-direction in cm.
    # The beam profile is assumed to be rectangular.
    input_configuration.addCard(Card("BEAM", what=["PROTON", -0.06, 0.0, 0.0, -2.0, -2.0], comment="beam source"))

    # add beam source position, see http://www.fluka.org/fluka.php?id=man_onl&sub=14
    beam_pos = Card("BEAMPOS", comment="beam source position")
    beam_pos.setWhat(1, 0.0)  # WHAT(1) = x-coordinate of the spot centre.
    beam_pos.setWhat(2, 0.0)  # WHAT(2) = y-coordinate of the spot centre.
    beam_pos.setWhat(3, -100.0)  # WHAT(3) = z-coordinate of the spot centre.
    input_configuration.addCard(beam_pos)

    # set geometry, in separate function
    set_geometry(input_configuration)

    # add scoring, in separate function
    add_scoring(input_configuration)

    # random number generator settings, see http://www.fluka.org/fluka.php?id=man_onl&sub=66
    rng_card = Card("RANDOMIZ")
    rng_card.setComment("random number generator settings")
    rng_card.setWhat(2, 137)  # Different numbers input as WHAT(2) will initialise different and independent
    # random number sequences, allowing the user to run several jobs in parallel for the same problem
    input_configuration.addCard(rng_card)

    # number of particles to simulate, see http://www.fluka.org/fluka.php?id=man_onl&sub=73
    start_card = Card("START")
    start_card.setComment("number of particles to simulate")
    start_card.setWhat(1, 1000)  # WHAT(1) - maximum number of primary histories simulated in the run
    input_configuration.addCard(start_card)

    # end of input configuration http://www.fluka.org/fluka.php?id=man_onl&sub=76
    input_configuration.addCard(Card("STOP"))

    # write file to the disk
    input_configuration.write(output_filename)

    return input_configuration


def set_geometry(input_file):
    """
    Add simple geometry, consisting of water target, surrounded with air.
    :param input_file: Fluka input file object, will be modified
    :return:
    """

    # geometry description starts here http://www.fluka.org/fluka.php?id=man_onl&sub=38
    geo_begin = Card("GEOBEGIN", comment="geometry description starts here")
    geo_begin.setSdum("COMBNAME")  # SDUM = COMBNAME  : Combinatorial geometry is used in free format,
    # and names can be used instead of body and region numbers
    input_file.addCard(geo_begin)

    # list of bodies, see http://www.fluka.org/fluka.php?id=man_onl&sub=94
    # 1. black body sphere, centered around (0,0,0), radius 10m
    blkbody_sph = Card("SPH", comment="black body sphere")
    blkbody_sph.setSdum("blkbody")  # SDUM - name of the body
    blkbody_sph.setWhat(1, 0.0)  # WHAT(1) - x-coordinate of the centre (in free format)
    blkbody_sph.setWhat(2, 0.0)  # WHAT(2) - y-coordinate of the centre (in free format)
    blkbody_sph.setWhat(3, 0.0)  # WHAT(3) - z-coordinate of the centre (in free format)
    blkbody_sph.setWhat(4, 10000.0)  # WHAT(4) - radius (in free format)
    input_file.addCard(blkbody_sph)

    # 2. air sphere, centered around (0,0,0), radius 1m
    air_sph = blkbody_sph.clone()  # see comments above
    air_sph.setSdum("air")
    air_sph.setWhat(4, 100.0)
    air_sph.setComment("air shpere")
    input_file.addCard(air_sph)

    # 3. target cylinder, bottom center at (0,0,0), height 10cm, radius 5cm, positioned along Z axis
    target_rcc = Card("RCC", comment="target cylinder")
    target_rcc.setSdum("target")  # SDUM - name of the body
    target_rcc.setWhat(1, 0.0)  # WHAT(1) - x-coordinate of the centre of circular plane, bottom (in free format)
    target_rcc.setWhat(2, 0.0)  # WHAT(2) - y-coordinate of the centre of circular plane, bottom (in free format)
    target_rcc.setWhat(3, 0.0)  # WHAT(3) - z-coordinate of the centre of circular plane, bottom (in free format)
    target_rcc.setWhat(4, 0.0)  # WHAT(4) - x-coordinate of the height vector (in free format)
    target_rcc.setWhat(5, 0.0)  # WHAT(5) - y-coordinate of the height vector (in free format)
    target_rcc.setWhat(6, 10.0)  # WHAT(6) - z-coordinate of the height vector (in free format)
    target_rcc.setWhat(7, 5.0)  # WHAT(7) - radius (in free format)
    input_file.addCard(target_rcc)

    # end of body region
    input_file.addCard(Card("END"))

    # list of regions, see http://www.fluka.org/fluka.php?id=man_onl&sub=94
    # 1. outer black body region
    blk_body_zone = Card("REGION")
    blk_body_zone.setComment("outer black body region")
    blk_body_zone.setSdum("Z_BBODY")  # SDUM - name of the region
    blk_body_zone.addZone('+blkbody -air')  # logical formula
    input_file.addCard(blk_body_zone)

    # 2. inner air region
    air_zone = Card("REGION")
    air_zone.setComment("inner air region")
    air_zone.setSdum("Z_AIR")
    air_zone.addZone('+air -target')
    input_file.addCard(air_zone)

    # 3. target region
    target_zone = Card("REGION")
    target_zone.setComment("target region")
    target_zone.setSdum("Z_TARGET")
    target_zone.addZone('+target')
    input_file.addCard(target_zone)

    # end of target region
    input_file.addCard(Card("END"))

    # geometry description ends here http://www.fluka.org/fluka.php?id=man_onl&sub=39
    geo_end = Card("GEOEND", comment="geometry description ends here")
    input_file.addCard(geo_end)

    # material assignment, see http://www.fluka.org/fluka.php?id=man_onl&sub=10
    blk_body_mat = Card("ASSIGNMA")
    blk_body_mat.setWhat(1, 'BLCKHOLE')  # WHAT(1) - material index, or material name
    blk_body_mat.setWhat(2, blk_body_zone.sdum())  # WHAT(2) - lower bound (or name corresponding to it) of the region
    # indices with material index equal or corresponding to WHAT(1).
    # upper bound is set to WHAT(2) by default
    input_file.addCard(blk_body_mat)

    air_mat = Card("ASSIGNMA")  # see comments above
    air_mat.setWhat(1, 'AIR')
    air_mat.setWhat(2, air_zone.sdum())
    input_file.addCard(air_mat)

    target_mat = Card("ASSIGNMA")  # see comments above
    target_mat.setWhat(1, 'WATER')
    target_mat.setWhat(2, target_zone.sdum())
    input_file.addCard(target_mat)


def add_scoring(input_file):
    """
    Add scoring of deposited energy and neutron fluence on 4 different mesh types (8 scorers in total)
    :param input_file: Fluka input_file object, will be modified
    :return:
    """

    # Dictionary of scoring type (generalized particle)
    # keys will be used as parts of output filenames, this is why they are short
    # for list of all possible generalized particles, see: http://www.fluka.org/fluka.php?id=man_onl&sub=7
    det_types = {"en": "ENERGY", "n": "NEUTRON"}

    # Define 4 different scoring meshes (1D and 2D), using pymchelper classes
    msh_z = CarthesianMesh()
    msh_z.set_axis(axis_no=0, start=-0.5, stop=0.5, nbins=1)
    msh_z.set_axis(axis_no=1, start=-0.5, stop=0.5, nbins=1)
    msh_z.set_axis(axis_no=2, start=0.0, stop=5.0, nbins=500)

    msh_xy = CarthesianMesh()
    msh_xy.set_axis(axis_no=0, start=-5.0, stop=5.0, nbins=500)
    msh_xy.set_axis(axis_no=1, start=-5.0, stop=5.0, nbins=500)
    msh_xy.set_axis(axis_no=2, start=2.8, stop=2.9, nbins=1)

    msh_yz = CarthesianMesh()
    msh_yz.set_axis(axis_no=0, start=-0.1, stop=0.1, nbins=1)
    msh_yz.set_axis(axis_no=1, start=-5.0, stop=5.0, nbins=500)
    msh_yz.set_axis(axis_no=2, start=0.0, stop=5.0, nbins=500)

    msh_zx = CarthesianMesh()
    msh_zx.set_axis(axis_no=0, start=-5.0, stop=5.0, nbins=500)
    msh_zx.set_axis(axis_no=1, start=-0.1, stop=0.1, nbins=1)
    msh_zx.set_axis(axis_no=2, start=0.0, stop=5.0, nbins=500)

    mesh_types = {"z": msh_z, "xy": msh_xy, "yz": msh_yz, "zx": msh_zx}

    # Fluka is using concept of output unit numbers, this is starting number,
    # for each new scorer will be decreased by one
    output_unit = -21

    # scoring, see http://www.fluka.org/fluka.php?id=man_onl&sub=84
    # template cards for USRBIN scorers:
    usrbin_card1 = Card("USRBIN")
    usrbin_card1.setWhat(1, 0.0)  # WHAT(1) - type of binning selected (0.0 : Mesh Cartesian, no sym.)
    usrbin_card2 = Card("USRBIN")
    usrbin_card2.setSdum("&")  # continuation card

    # loop over all possible detectors and meshes
    for det_type, mesh_type in product(det_types.keys(), mesh_types.keys()):

        # make copies of template cards
        card1 = usrbin_card1.clone()
        card2 = usrbin_card2.clone()

        # generate comment
        card1.setComment("scoring {:s} on mesh {:s}".format(det_types[det_type], mesh_type))
        card1.setSdum("{:s}_{:s}".format(det_type, mesh_type))  # SDUM - identifying the binning detector
        card1.setWhat(2, det_types[det_type])  # WHAT(2) - particle (or particle family) type to be scored
        card1.setWhat(3, output_unit)  # WHAT(3) - output unit (> 0.0 : formatted data,  < 0.0 : unformatted data)

        msh = mesh_types[mesh_type]
        card1.setWhat(4, msh.axis[0].stop)  # WHAT(4) - For Cartesian binning: Xmax
        card1.setWhat(5, msh.axis[1].stop)  # WHAT(5) - For Cartesian binning: Ymax
        card1.setWhat(6, msh.axis[2].stop)  # WHAT(6) - For Cartesian binning: Zmax
        card2.setWhat(1, msh.axis[0].start)  # WHAT(1) - For Cartesian binning: Xmin
        card2.setWhat(2, msh.axis[1].start)  # WHAT(2) - For Cartesian binning: Ymin
        card2.setWhat(3, msh.axis[2].start)  # WHAT(3) - For Cartesian binning: Zmin
        card2.setWhat(4, msh.axis[0].nbins)  # WHAT(4) - For Cartesian binning: number of X bins
        card2.setWhat(5, msh.axis[1].nbins)  # WHAT(5) - For Cartesian binning: number of Y bins
        card2.setWhat(6, msh.axis[2].nbins)  # WHAT(6) - For Cartesian binning: number of Z bins
        output_unit -= 1

        # finally add generated cards to input structure
        input_file.addCard(card1)
        card1.appendWhats(card2.whats()[1:])


def main(args=sys.argv[1:]):

    generate_fluka_file("fl_sim.inp")

    # fl_sim.inp file should have following content:
    # TITLE
    # proton beam simulation
    # * default physics settings for hadron therapy
    # DEFAULTS                                                              HADROTHE
    # * beam source
    # BEAM           -0.06       0.0       0.0      -2.0      -2.0          PROTON
    # * beam source position
    # BEAMPOS          0.0       0.0    -100.0
    # * geometry description starts here
    # GEOBEGIN                                                              COMBNAME
    #     0    0
    # * black body sphere
    # SPH blkbody    0.0 0.0 0.0 10000.0
    # * air shpere
    # SPH air        0.0 0.0 0.0 100.0
    # * target cylinder
    # RCC target     0.0 0.0 0.0 0.0 0.0 10.0 5.0
    # END
    # * outer black body region
    # Z_BBODY      5 +blkbody -air
    # * inner air region
    # Z_AIR        5 +air -target
    # * target region
    # Z_TARGET     5 +target
    # END
    # GEOEND
    # ASSIGNMA    BLCKHOLE   Z_BBODY
    # ASSIGNMA         AIR     Z_AIR
    # ASSIGNMA       WATER  Z_TARGET
    # * scoring NEUTRON on mesh z
    # USRBIN           0.0   NEUTRON       -21       0.5       0.5       5.0n_z
    # USRBIN          -0.5      -0.5       0.0         1         1       500&
    # * scoring NEUTRON on mesh yz
    # USRBIN           0.0   NEUTRON       -22       0.1       5.0       5.0n_yz
    # USRBIN          -0.1      -5.0       0.0         1       500       500&
    # * scoring NEUTRON on mesh xy
    # USRBIN           0.0   NEUTRON       -23       5.0       5.0       2.9n_xy
    # USRBIN          -5.0      -5.0       2.8       500       500         1&
    # * scoring NEUTRON on mesh zx
    # USRBIN           0.0   NEUTRON       -24       5.0       0.1       5.0n_zx
    # USRBIN          -5.0      -0.1       0.0       500         1       500&
    # * scoring ENERGY on mesh z
    # USRBIN           0.0    ENERGY       -25       0.5       0.5       5.0en_z
    # USRBIN          -0.5      -0.5       0.0         1         1       500&
    # * scoring ENERGY on mesh yz
    # USRBIN           0.0    ENERGY       -26       0.1       5.0       5.0en_yz
    # USRBIN          -0.1      -5.0       0.0         1       500       500&
    # * scoring ENERGY on mesh xy
    # USRBIN           0.0    ENERGY       -27       5.0       5.0       2.9en_xy
    # USRBIN          -5.0      -5.0       2.8       500       500         1&
    # * scoring ENERGY on mesh zx
    # USRBIN           0.0    ENERGY       -28       5.0       0.1       5.0en_zx
    # USRBIN          -5.0      -0.1       0.0       500         1       500&
    # * random number generator settings
    # RANDOMIZ                   137
    # * number of particles to simulate
    # START           1000
    # STOP


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



import setuptools

from pymchelper.version import git_version


def write_version_py(filename='VERSION'):
    cnt = """%(version)s
"""

    GIT_REVISION = git_version()
    a = open(filename, 'w')
    try:
        a.write(cnt % {'version': GIT_REVISION})
    finally:
        a.close()


write_version_py()

with open('README.rst') as readme_file:
    readme = readme_file.read()


setuptools.setup(
    name='pymchelper',
    version=git_version(),
    packages=setuptools.find_packages(where='.', exclude=("*.tests", "*.tests.*", "tests.*", "tests")),
    url='https://github.com/DataMedSci/pymchelper',
    license='MIT',
    author='Leszek Grzanka',
    author_email='leszek.grzanka@gmail.com',
    description='Python toolkit for SHIELD-HIT12A and Fluka',
    long_description=readme + '\n',
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Scientific/Engineering :: Physics',

        # OS and env
        'Environment :: Console',
        'Operating System :: POSIX :: Linux',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    entry_points={
        'console_scripts': [
            'convertmc=pymchelper.run:main',
            'pld2sobp=pymchelper.utils.pld2sobp:main',
            'mcscripter=pymchelper.utils.mcscripter:main',
        ],
    },
    package_data={'pymchelper': ['flair/db/*']},
    install_requires=[
        'enum34',
        'numpy'
    ],
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.3.*',
)



# $Id: rexx.py 3312 2014-10-17 07:25:38Z bnv $
#
# Copyright and User License
# ~~~~~~~~~~~~~~~~~~~~~~~~~~
# Copyright Vasilis.Vlachoudis@cern.ch for the
# European Organization for Nuclear Research (CERN)
#
# All rights not expressly granted under this license are reserved.
#
# Installation, use, reproduction, display of the
# software ("flair"), in source and binary forms, are
# permitted free of charge on a non-exclusive basis for
# internal scientific, non-commercial and non-weapon-related
# use by non-profit organizations only.
#
# For commercial use of the software, please contact the main
# author Vasilis.Vlachoudis@cern.ch for further information.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the
#    distribution.
#
# DISCLAIMER
# ~~~~~~~~~~
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
# NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY, OF
# SATISFACTORY QUALITY, AND FITNESS FOR A PARTICULAR PURPOSE
# OR USE ARE DISCLAIMED. THE COPYRIGHT HOLDERS AND THE
# AUTHORS MAKE NO REPRESENTATION THAT THE SOFTWARE AND
# MODIFICATIONS THEREOF, WILL NOT INFRINGE ANY PATENT,
# COPYRIGHT, TRADE SECRET OR OTHER PROPRIETARY RIGHT.
#
# LIMITATION OF LIABILITY
# ~~~~~~~~~~~~~~~~~~~~~~~
# THE COPYRIGHT HOLDERS AND THE AUTHORS SHALL HAVE NO
# LIABILITY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL,
# CONSEQUENTIAL, EXEMPLARY, OR PUNITIVE DAMAGES OF ANY
# CHARACTER INCLUDING, WITHOUT LIMITATION, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES, LOSS OF USE, DATA OR PROFITS,
# OR BUSINESS INTERRUPTION, HOWEVER CAUSED AND ON ANY THEORY
# OF CONTRACT, WARRANTY, TORT (INCLUDING NEGLIGENCE), PRODUCT
# LIABILITY OR OTHERWISE, ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
#
# Author:	Vasilis.Vlachoudis@cern.ch
# Date:	14-May-2004

import string

__author__ = "Vasilis Vlachoudis"
__email__ = "Vasilis.Vlachoudis@cern.ch"

_letters_digits = string.ascii_letters + string.digits
_letters_digits_symbol = _letters_digits + "_."


# abbrev
def abbrev(information, info, length=0):
    """
    return true if the info is an abbreviation of information
    with minimum length l
    """
    if length > 0:
        length = length
    else:
        length = len(info)

    cond1 = (len(information) >= len(info))
    cond2 = (len(info) >= length)
    cond3 = (information[:len(info)] == info)
    return cond1 and cond2 and cond3


# center
def center(s, length, pad=' '):
    if length <= 0:
        return ""

    i = length - len(s)
    if i == 0:
        return s
    elif i < 0:
        i = -i
        a = i // 2
        return s[a:a + length]
    else:
        a = i // 2
        return "%s%s%s" % (pad * a, s, pad * (i - a))


# datatype
def datatype(str, check="N"):
    """rexx datatype function"""

    try:
        if len(str) == 0:
            return check == "X" or check == "B"
    except Exception:
        return check == "X" or check == "B"

    if check == "N":
        return _isnum(str)

    if check == "A":
        return verify(str, _letters_digits) == -1
    elif check == "L":
        return verify(str, string.ascii_lowercase) == -1
    elif check == "M":
        return verify(str, string.ascii_letters) == -1
    elif check == "U":
        return verify(str, string.ascii_uppercase) == -1
    elif check == "O":
        return verify(str, string.octdigits) == -1
    elif check == "X":
        return verify(str, string.hexdigits) == -1
    elif check == "S":
        return (str[0] in string.ascii_letters) and \
               (verify(str[1:], _letters_digits_symbol) == -1)
    else:
        return _isnum(str)


# insert
def insert(new, target, n, pad=" "):
    """
    insert new string to target as position n padded with pad characters
    """
    if n == 0:
        return new + target
    elif n > len(target):
        return target + pad * (n - len(target)) + new

    return target[0:n] + new + target[n:]


# left
def left(str, length, pad=" "):
    """return left of string str of length padded with pad chars"""
    if length < len(str):
        return str[0:length]
    else:
        return str + (pad * (length - len(str)))


# translate
def translate(str, tableo=None, tablei=None, pad=" "):
    """translate string"""
    # If neither input nor output tables, uppercase.
    if tableo is None and tablei is None:
        return str.upper()

    if tableo is None:
        tableo = xrange_string(0, 255)

    if tablei is None:
        tablei = xrange_string(0, 255)

    # The input table defaults to all characters.
    dl = len(tablei) - len(tableo)
    if dl > 0:
        tableo += pad * dl
    else:
        tablei += pad * (-dl)

    tbl = string.maketrans(tablei, tableo)
    return str.translate(tbl)


# reverse
def reverse(str):
    """reverse string"""
    return str[::-1]


# verify
def verify(str, ref, match=0, start=0):
    """
    return the index of the first character in string that
    is not also in reference. if "Match" is given, then return
    the result index of the first character in string that is in reference
    """

    if start < 0:
        start = 0
    if start >= len(str):
        return -1

    for i in range(start, len(str)):
        found = ref.find(str[i]) == -1
        if found ^ match:
            return i
    return -1


# xrange
def xrange_string(start, stop):
    return "".join([chr(x) for x in range(start, stop + 1)])


def _isnum(str):
    """true if string is number"""
    str = str.strip()

    # accept one sign
    i = 0
    length = len(str)

    if length == 0:
        return False

    if str[i] == '-' or str[i] == '+':
        i += 1

    # skip spaces after sign
    while i < length and str[i].isspace():
        i += 1

    # accept many digits
    if i < length and '0' <= str[i] <= '9':
        i += 1
        F = 1
        while i < length and '0' <= str[i] <= '9':
            i += 1
    else:
        F = 0

    # accept one dot
    if i < length and str[i] == '.':
        i += 1

        # accept many digits
        if i < length and '0' <= str[i] <= '9':
            while i < length and '0' <= str[i] <= '9':
                i += 1
        else:
            if not F:
                return False
    else:
        if not F:
            return False

    # accept one e/E/d/D
    if i < length and (str[i] == 'e' or str[i] == 'E' or str[i] == 'd' or str[i] == 'D'):
        i += 1
        # accept one sign
        if i < length and (str[i] == '-' or str[i] == '+'):
            i += 1

        # accept many digits
        if i < length and '0' <= str[i] <= '9':
            while i < length and '0' <= str[i] <= '9':
                i += 1
        else:
            return False

    if i != length:
        return False

    return True


if __name__ == "__main__":
    from pymchelper.flair.common.log import say

    say("abbrev")
    assert abbrev('information', 'info', 4)
    assert abbrev('information', '', 0)
    assert not abbrev('information', 'Info', 4)
    assert not abbrev('information', 'info', 5)
    assert not abbrev('information', 'info ')
    assert abbrev('information', 'info', 3)
    assert not abbrev('info', 'information', 3)
    assert not abbrev('info', 'info', 5)

    say("center")
    assert center('****', 0, '-') == ''
    assert center('****', 8, '-') == '--****--'
    assert center('****', 7, '-') == '-****--'
    assert center('*****', 8, '-') == '-*****--'
    assert center('*****', 7, '-') == '-*****-'
    assert center('12345678', 4, '-') == '3456'
    assert center('12345678', 5, '-') == '23456'
    assert center('1234567', 4, '-') == '2345'
    assert center('1234567', 5, '-') == '23456'

    say("datatype")
    assert not datatype("")
    assert not datatype("foobar")
    assert not datatype("foo bar")
    assert not datatype("123.456.789")
    assert datatype("123.456")
    assert not datatype("DeadBeef")
    assert not datatype("Dead Beef")
    assert not datatype("1234ABCD")
    assert datatype("01001101")
    assert not datatype("0110 1101")
    assert not datatype("0110 101")
    assert datatype("1324.1234")
    assert datatype("123")
    assert datatype("12.3")
    assert datatype('123.123')
    assert datatype('123.123E3')
    assert datatype('123.0000003')
    assert datatype('123.0000004')
    assert datatype('123.0000005')
    assert datatype('123.0000006')
    assert datatype(' 23')
    assert datatype(' 23 ')
    assert datatype('23 ')
    assert datatype('123.00')
    assert datatype('123000E-2')
    assert datatype('123000E+2')
    assert not datatype("A B C")
    assert not datatype("123ABC")
    assert not datatype("123AHC")
    assert datatype('0.000E-2')
    assert datatype('0.000E-1')
    assert datatype('0.000E0')
    assert datatype('0.000E1')
    assert datatype('0.000E2')
    assert datatype('0.000E3')
    assert datatype('0.000E4')
    assert datatype('0.000E5')
    assert datatype('0.000E6')
    assert datatype('0E-1')
    assert datatype('0E0')
    assert datatype('0E1')
    assert datatype('0E2')
    assert not datatype('+.')
    assert not datatype('++0')

    say("insert")
    assert insert("abc", "def", 2) == "deabcf"
    assert insert("abc", "def", 3) == "defabc"
    assert insert("abc", "def", 5) == "def  abc"
    assert insert("abc", "def", 5, '*') == "def**abc"

    say("translate")
    assert translate("Foo Bar", "", "") == "Foo Bar"
    assert translate("Foo Bar", xrange_string(1, 255)) == "Gpp!Cbs"
    assert translate("", "klasjdf", "woieruw") == ""
    assert translate("foobar", "abcdef", "fedcba") == "aooefr"

    say("verify")
    assert verify('foobar', 'barfo', 0, 0) == -1
    assert verify('foobar', 'barfo', 1, 0) == 0
    assert verify('', 'barfo') == -1
    assert verify('foobar', '') == 0
    assert verify('foobar', 'barf', 0, 2) == 2
    assert verify('foobar', 'barf', 0, 3) == -1
    assert verify('', '') == -1

    say("All Test passed")



# Python2 lacks a FileNotFoundError exception, we use then IOError
try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError

try:
    with open('VERSION', 'r') as f:
        __version__ = f.readline(1)
except FileNotFoundError:
    from pymchelper.version import git_version
    __version__ = git_version()



import logging
import math
import os
import time
import numpy as np

logger = logging.getLogger(__name__)


class TripCubeWriter:
    def __init__(self, filename, options):
        self.output_corename = filename

    def write(self, detector):
        import getpass
        from pymchelper.shieldhit.detector.detector_type import SHDetType
        from pymchelper import __version__ as _pmcversion
        try:
            from pytrip import __version__ as _ptversion
        except ImportError:
            logger.error("pytrip package missing, to install type `pip install pytrip98`")
            return 1

        pixel_size_x = (detector.x.max_val - detector.x.min_val) / detector.x.n
        pixel_size_z = (detector.z.max_val - detector.z.min_val) / detector.z.n

        logging.debug("psx: {:.6f} [cm]".format(pixel_size_x))
        logging.debug("psz: {:.6f} [cm]".format(pixel_size_z))

        _patient_name = "Anonymous"
        _created_by = getpass.getuser()
        _creation_info = "Created with pymchelper {:s}; using PyTRiP98 {:s}".format(_pmcversion,
                                                                                    _ptversion)

        if detector.dettyp == SHDetType.dose:

            from pytrip import dos

            cube = dos.DosCube()
            # Warning: PyTRiP cube dimensions are in [mm]
            cube.create_empty_cube(
                1.0, detector.x.n, detector.y.n, detector.z.n,
                pixel_size=pixel_size_x * 10.0,
                slice_distance=pixel_size_z * 10.0)

            # .dos dose cubes are usually in normalized integers,
            # where "1000" equals 100.0 % dose.
            # The next are also the defaults, but just to be clear
            # this is specifically set.
            cube.data_type = "integer"
            cube.num_bytes = 2
            cube.pydata_type = np.int16

            cube.cube = detector.data

            if detector.tripdose >= 0.0 and detector.tripntot > 0:
                cube.cube = (cube.cube * detector.tripntot * 1.602e-10) / detector.tripdose * 1000.0
            else:
                cube.cube = (cube.cube / cube.cube.max()) * 1200.0

            # Save proper meta information
            cube.patient_name = _patient_name
            cube.created_by = _created_by
            cube.creation_info = _creation_info

            cube.write(self.output_corename)

            return 0

        elif detector.dettyp in (SHDetType.dlet, SHDetType.tlet, SHDetType.dletg, SHDetType.tletg):

            from pytrip import let

            cube = let.LETCube()
            # Warning: PyTRiP cube dimensions are in [mm]
            cube.create_empty_cube(
                1.0, detector.x.n, detector.y.n, detector.z.n,
                pixel_size=pixel_size_x * 10.0,
                slice_distance=pixel_size_z * 10.0)

            # .dosemlet.dos LET cubes are usually in 32 bit floats.
            cube.data_type = "float"
            cube.num_bytes = 4
            cube.pydata_type = np.float32

            # need to redo the cube, since by default np.float32 are allocated.
            # When https://github.com/pytrip/pytrip/issues/35 is fixed,
            # then this should not be needed.
            cube.cube = np.ones((cube.dimz, cube.dimy, cube.dimx), dtype=cube.pydata_type)

            cube.cube = detector.data
            cube.cube *= 0.1  # MeV/cm -> keV/um
            # Save proper meta information

            cube.patient_name = _patient_name
            cube.created_by = _created_by
            cube.creation_info = _creation_info

            cube.write(self.output_corename)

            return 0

        else:
            logger.error("Tripcube target is only allowed with dose- or LET-type detectors.")
            raise Exception("Illegal detector for tripcube.")

            return 1


class TripDddWriter(object):

    _sigma_to_fwhm = 2. * (2. * math.log(2.)) ** 0.5

    _ddd_header_template = """!filetype    ddd
!fileversion   {fileversion:s}
!filedate      {filedate:s}
!projectile    {projectile:s}
!material      {material:s}
!composition   {composition:s}
!density {density:f}
!energy {energy:f}
#
# {creator:s}
#
#   z[g/cm**2] dE/dz[MeV/(g/cm**2)] FWHM1[g/cm**2] factor FWHM2[g/cm**2]
!ddd
"""

    def __init__(self, filename, options):

        import matplotlib
        matplotlib.use('Agg')
        self.ddd_filename = filename
        self.energy_MeV = options.energy
        self.projectile = options.projectile
        self.ngauss = options.ngauss
        self.verbosity = options.verbose
        if not self.ddd_filename.endswith(".ddd"):
            self.ddd_filename += ".ddd"
        self.outputdir = os.path.abspath(os.path.dirname(self.ddd_filename))
        self.threshold = 3e-3
        env_var_name = 'DDD_TAIL_THRESHOLD'
        if env_var_name in os.environ:
            self.threshold = float(os.environ[env_var_name])
            logger.info("Setting tail threshold based on {:s} to {:f}".format(env_var_name, self.threshold))

    def write(self, detector):
        from pymchelper.shieldhit.detector.detector_type import SHDetType

        if detector.dettyp != SHDetType.ddd:
            logger.warning("Incompatible detector type {:s} used, please use {:s} instead".format(
                detector.dettyp, SHDetType.ddd))
            return 1

        # guess projectile and energy
        if self.projectile == '0':
            try:
                element_names = ['n', 'H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na', 'Mg', 'Al', 'Si',
                                 'P', 'S', 'Cl', 'Ar', 'K', 'Ca', 'Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu',
                                 'Zn', 'Ga', 'Ge', 'As', 'Se', 'Br', 'Kr', 'Rb', 'Sr', 'Y', 'Zr', 'Nb', 'Mo', 'Tc',
                                 'Ru', 'Rh', 'Pd', 'Ag', 'Cd', 'In', 'Sn', 'Sb', 'Te', 'I', 'Xe', 'Cs', 'Ba', 'La',
                                 'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho', 'Er', 'Tm', 'Yb', 'Lu',
                                 'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir', 'Pt', 'Au', 'Hg', 'Tl', 'Pb', 'Bi', 'Po', 'At',
                                 'Rn', 'Fr', 'Ra', 'Ac', 'Th', 'Pa', 'U', 'Np', 'Pu', 'Am', 'Cm', 'Bk', 'Cf', 'Es',
                                 'Fm', 'Md', 'No', 'Lr', 'Rf', 'Db', 'Sg', 'Bh', 'Hs', 'Mt', 'Ds', 'Rg', 'Cn', 'Nh',
                                 'Fl', 'Mc', 'Lv', 'Ts', 'Og']
                self.projectile = "{:d}{:s}".format(
                    int(detector.projectile_a),
                    element_names[int(detector.projectile_z)],
                )
            except AttributeError:
                logger.error('Projectile energy not available in raw_data, setting to 0')

        if self.energy_MeV == 0:
            try:
                self.energy_MeV = detector.projectile_energy
            except AttributeError:
                logger.error('Projectile energy not available in raw_data, setting to 0')

        # extract data from detector data
        self._extract_data(detector)

        # in order to avoid fitting data to noisy region far behind Bragg peak tail,
        # find the range of z coordinate which containes (1-threshold) of the deposited energy
        cum_dose = self._cumulative_dose()
        cum_dose_left = self._cumulative_dose_left(cum_dose)

        thr_ind = cum_dose_left.size - np.searchsorted(cum_dose_left[::-1], self.threshold) - 1
        z_fitting_cm_1d = self.z_data_cm_1d[:thr_ind]
        dose_fitting_MeV_g_1d = self.dose_data_MeV_g_1d[:thr_ind]

        r_fitting_cm_2d, z_fitting_cm_2d = np.meshgrid(self.r_data_cm_1d, z_fitting_cm_1d)
        dose_fitting_MeV_g_2d = self.dose_data_MeV_g_2d[0:thr_ind]

        logger.info("Plotting 1..")
        if self.verbosity > 0:
            self._pre_fitting_plots(
                cum_dose_left=cum_dose_left,
                z_fitting_cm_1d=z_fitting_cm_1d,
                dose_fitting_MeV_g_1d=dose_fitting_MeV_g_1d,
                threshold=self.threshold,
                zmax_cm=z_fitting_cm_1d[-1])

            self._plot_2d_map(z_fitting_cm_2d, r_fitting_cm_2d, dose_fitting_MeV_g_2d, z_fitting_cm_1d)

        logger.info("Fitting...")
        fwhm1_cm_data = np.zeros_like(z_fitting_cm_1d)
        fwhm2_cm_data = np.zeros_like(z_fitting_cm_1d)
        weight_data = np.zeros_like(z_fitting_cm_1d)
        dz0_MeV_cm_g_data = np.zeros_like(z_fitting_cm_1d)
        fwhm1_cm_error_data = np.zeros_like(z_fitting_cm_1d)
        fwhm2_cm_error_data = np.zeros_like(z_fitting_cm_1d)
        weight_error_data = np.zeros_like(z_fitting_cm_1d)
        dz0_MeV_cm_g_error_data = np.zeros_like(z_fitting_cm_1d)
        if self.ngauss in (1, 2):
            # for each depth fit a lateral beam with gaussian models
            for ind, z_cm in enumerate(z_fitting_cm_1d):

                dose_at_z = self.dose_data_MeV_g_2d[ind]

                # take into account only this position in r for which dose is positive
                r_fitting_cm = self.r_data_cm_1d[dose_at_z > 0]
                dose_fitting_1d_positive_MeV_g = dose_at_z[dose_at_z > 0]

                # perform the fit
                params, params_error = self._lateral_fit(r_fitting_cm,
                                                         dose_fitting_1d_positive_MeV_g,
                                                         self.ngauss)

                fwhm1_cm, factor, fwhm2_cm, dz0_MeV_cm_g = params
                fwhm1_cm_error, factor_error, fwhm2_cm_error, dz0_MeV_cm_g_error = params_error
                fwhm1_cm_data[ind] = fwhm1_cm
                dz0_MeV_cm_g_data[ind] = dz0_MeV_cm_g
                fwhm1_cm_error_data[ind] = fwhm1_cm_error
                dz0_MeV_cm_g_error_data[ind] = dz0_MeV_cm_g_error
                if self.ngauss == 2:
                    fwhm2_cm_data[ind] = fwhm2_cm  # set to 0 in case ngauss = 1
                    weight_data[ind] = factor  # set to 0 in case ngauss = 1
                    fwhm2_cm_error_data[ind] = fwhm2_cm_error
                    weight_error_data[ind] = factor_error

        logger.info("Plotting 2...")
        if self.verbosity > 0 and self.ngauss in (1, 2):
            self._post_fitting_plots(z_fitting_cm_1d,
                                     dose_fitting_MeV_g_1d,
                                     dz0_MeV_cm_g_data,
                                     fwhm1_cm_data,
                                     fwhm2_cm_data,
                                     weight_data,
                                     dz0_MeV_cm_g_error_data,
                                     fwhm1_cm_error_data,
                                     weight_error_data,
                                     fwhm2_cm_error_data)
            self._plot_2d_map(
                z_fitting_cm_2d,
                r_fitting_cm_2d,
                dose_fitting_MeV_g_2d,
                z_fitting_cm_1d,
                fwhm1_cm_data,
                fwhm2_cm_data,
                weight_data,
                dz0_MeV_cm_g_data,
                suffix='_fwhm')

        logger.info("Writing " + self.ddd_filename)

        from pymchelper import __version__ as _pmcversion
        try:
            from pytrip import __version__ as _ptversion
        except ImportError:
            logger.error("pytrip package missing, to install type `pip install pytrip98`")
            return 1
        _creator_info = "Created with pymchelper {:s}; using PyTRiP98 {:s}".format(_pmcversion,
                                                                                   _ptversion)

        # prepare header of DDD file
        header = self._ddd_header_template.format(
            fileversion='19980520',
            filedate=time.strftime('%c'),  # Locale's appropriate date and time representation
            projectile=self.projectile,
            material='H2O',
            composition='H2O',
            density=1,
            creator=_creator_info,
            energy=self.energy_MeV)

        # write the contents of the files
        with open(self.ddd_filename, 'w') as ddd_file:
            ddd_file.write(header)
            # TODO write to DDD gaussian amplitude, not the dose in central bin
            if self.ngauss == 2:
                for z_cm, dose, fwhm1_cm, weight, fwhm2_cm in zip(z_fitting_cm_1d, dose_fitting_MeV_g_1d,
                                                                  fwhm1_cm_data, weight_data, fwhm2_cm_data):
                    ddd_file.write('{:g} {:g} {:g} {:g} {:g}\n'.format(z_cm, dose, fwhm1_cm, weight, fwhm2_cm))
            elif self.ngauss == 1:
                for z_cm, dose, fwhm_cm in zip(z_fitting_cm_1d, dose_fitting_MeV_g_1d, fwhm1_cm_data):
                    ddd_file.write('{:g} {:g} {:g}\n'.format(z_cm, dose, fwhm_cm))
            elif self.ngauss == 0:
                for z_cm, dose in zip(z_fitting_cm_1d, dose_fitting_MeV_g_1d):
                    ddd_file.write('{:g} {:g}\n'.format(z_cm, dose))

        return 0

    def _extract_data(self, detector):
        # 1D arrays of r,z
        self.r_data_cm_1d = detector.x.data
        self.z_data_cm_1d = detector.z.data

        # 2D arrays of r,z, dose and error
        self.r_data_cm_2d, self.z_data_cm_2d = np.meshgrid(self.r_data_cm_1d, self.z_data_cm_1d)

        self.dose_data_MeV_g_2d = detector.data_raw.reshape((detector.z.n, detector.x.n))
        self.dose_error_MeV_g_2d = detector.error_raw.reshape((detector.z.n, detector.x.n))

        # dose in the very central bin
        bin_depth_z_cm = self.z_data_cm_1d[1] - self.z_data_cm_1d[0]
        r_step_cm = self.r_data_cm_1d[1] - self.r_data_cm_1d[0]

        # Bin volume increases as we move away from beam axis
        # i-th bin volume = dz * pi * (r_i_max^2 - r_i_min^2  )
        #   r_i_max = r_i + dr / 2
        #   r_i_min = r_i - dr / 2
        #  r_i_max^2 - r_i_min^2 = (r_i_max - r_i_min)*(r_i_max + r_i_min) = dr * 2 * r_i    thus
        # i-th bin volume = 2 * pi * dr * r_i * dz
        bin_volume_data_cm3_1d = 2.0 * np.pi * r_step_cm * self.r_data_cm_1d * bin_depth_z_cm
        # we assume density of 1 g/c3
        density_g_cm3 = 1.0
        energy_in_bin_MeV_2d = self.dose_data_MeV_g_2d * bin_volume_data_cm3_1d * density_g_cm3
        total_bin_mass_g = density_g_cm3 * bin_depth_z_cm * np.pi * (self.r_data_cm_1d[-1] + r_step_cm / 2.0)**2
        total_energy_at_depth_MeV_1d = np.sum(energy_in_bin_MeV_2d, axis=1)
        self.dose_data_MeV_g_1d = total_energy_at_depth_MeV_1d / total_bin_mass_g

    def _cumulative_dose(self):
        cumsum = np.cumsum(self.dose_data_MeV_g_1d)
        cumsum /= np.sum(self.dose_data_MeV_g_1d)
        return cumsum

    def _cumulative_dose_left(self, cumsum):
        cum_dose_left = np.array(cumsum)
        cum_dose_left *= -1.0
        cum_dose_left += 1.0
        return cum_dose_left

    def _plot_2d_map(self,
                     z_fitting_cm_2d,
                     r_fitting_cm_2d,
                     dose_fitting_MeV_g2d,
                     z_fitting_cm_1d=None,
                     fwhm1_cm=None,
                     fwhm2_cm=None,
                     weight=None,
                     dz0_MeV_cm_g_data=None,
                     suffix=''):
        import matplotlib
        matplotlib.use('Agg')
        import matplotlib.pyplot as plt
        from matplotlib.colors import LogNorm

        prefix = os.path.join(self.outputdir,
                              '{:s}_plot_{:3.1f}MeV_'.format(
                                  os.path.splitext(os.path.basename(self.ddd_filename))[0],
                                  self.energy_MeV))

        plt.pcolormesh(z_fitting_cm_2d, r_fitting_cm_2d, dose_fitting_MeV_g2d,
                       norm=LogNorm(), cmap='gnuplot2', label='dose')
        cbar = plt.colorbar()
        cbar.set_label("dose [MeV/g]", rotation=270, verticalalignment='bottom')
        if z_fitting_cm_1d is not None and np.any(fwhm1_cm):
            plt.plot(z_fitting_cm_1d, fwhm1_cm, color='g', label="fwhm1")
        if z_fitting_cm_1d is not None and np.any(fwhm2_cm):
            plt.plot(z_fitting_cm_1d, fwhm2_cm, color='r', label="fwhm2")

        # plot legend only if some of the FWHM 1-D overlays are present
        # adding legend to only pcolormesh plot will result in a warning about missing labels
        if z_fitting_cm_1d is not None and (np.any(fwhm1_cm) or np.any(fwhm2_cm)):
            plt.legend(loc=0)
        plt.xlabel("z [cm]")
        plt.ylabel("r [cm]")
        plt.xlim((z_fitting_cm_2d.min(), z_fitting_cm_2d.max()))
        if np.any(fwhm1_cm) and np.any(fwhm2_cm):
            plt.ylim((r_fitting_cm_2d.min(), max(max(fwhm1_cm), max(fwhm2_cm))))
        plt.clim((1e-8 * dose_fitting_MeV_g2d.max(), dose_fitting_MeV_g2d.max()))
        out_filename = prefix + 'dosemap' + suffix + '.png'
        logger.info('Saving ' + out_filename)
        plt.savefig(out_filename)
        if self.verbosity > 1:
            plt.yscale('log')
            out_filename = prefix + 'dosemap_log' + suffix + '.png'
            logger.info('Saving ' + out_filename)
            plt.savefig(out_filename)
        plt.close()

        if self.verbosity > 2 and (np.any(fwhm1_cm) or np.any(fwhm2_cm)):
            # TODO add plotting sum of 2 gausses

            sigma1_cm = fwhm1_cm / self._sigma_to_fwhm
            sigma2_cm = fwhm2_cm / self._sigma_to_fwhm
            gauss_amplitude_MeV_g = dz0_MeV_cm_g_data
            for z_cm, sigma1_at_z_cm, sigma2_at_z_cm, factor, amplitude_MeV_g in \
                    zip(z_fitting_cm_1d, sigma1_cm, sigma2_cm, weight, gauss_amplitude_MeV_g):
                dose_mc_MeV_g = self.dose_data_MeV_g_2d[self.z_data_cm_2d == z_cm]
                title = "Z = {:4.3f} cm,  sigma1 = {:4.3f} cm".format(z_cm, sigma1_at_z_cm)
                plt.plot(self.r_data_cm_1d, dose_mc_MeV_g, 'k.', label="data")
                if self.ngauss == 1:
                    gauss_data_MeV_g = self.gauss_MeV_g(self.r_data_cm_1d, amplitude_MeV_g, sigma1_at_z_cm)
                    plt.plot(self.r_data_cm_1d, gauss_data_MeV_g, label="fit")
                elif self.ngauss == 2:
                    gauss_data_MeV_g = self.gauss2_MeV_g(self.r_data_cm_1d, amplitude_MeV_g,
                                                         sigma1_at_z_cm, factor, sigma2_at_z_cm)
                    gauss_data_MeV_g_1st = self.gauss2_MeV_g_1st(self.r_data_cm_1d, amplitude_MeV_g,
                                                                 sigma1_at_z_cm, factor, sigma2_at_z_cm)
                    gauss_data_MeV_g_2nd = self.gauss2_MeV_g_2nd(self.r_data_cm_1d, amplitude_MeV_g,
                                                                 sigma1_at_z_cm, factor, sigma2_at_z_cm)
                    plt.plot(self.r_data_cm_1d, gauss_data_MeV_g, label="fit")
                    plt.plot(self.r_data_cm_1d, gauss_data_MeV_g_1st, label="fit 1st gauss")
                    plt.plot(self.r_data_cm_1d, gauss_data_MeV_g_2nd, label="fit 2nd gauss")
                    title += ", sigma2 = {:4.3f} cm, factor = {:4.6f}".format(sigma2_at_z_cm, factor)
                logger.debug("Plotting at " + title)
                plt.title(title)
                plt.legend(loc=0)
                plt.yscale('log')
                plt.xlabel("r [cm]")
                plt.ylabel("dose [MeV/g]")
                plt.ylim([dose_mc_MeV_g.min(), dose_mc_MeV_g.max()])
                if self.ngauss == 1:
                    plt.ylim([dose_mc_MeV_g.min(), max(gauss_data_MeV_g.max(), dose_mc_MeV_g.max())])
                out_filename = prefix + "fit_details_{:4.3f}_log".format(z_cm) + suffix + '.png'
                logger.info('Saving ' + out_filename)
                plt.savefig(out_filename)

                plt.xscale('log')
                plt.xlim([0, self.r_data_cm_1d.max()])
                out_filename = prefix + "fit_details_{:4.3f}_loglog".format(z_cm) + suffix + '.png'
                logger.info('Saving ' + out_filename)
                plt.savefig(out_filename)

                plt.xscale('linear')
                plt.xlim([0, 5.0 * sigma2_at_z_cm])
                plt.ylim([dose_mc_MeV_g[self.r_data_cm_1d < 5.0 * sigma2_at_z_cm].min(), dose_mc_MeV_g.max()])
                out_filename = prefix + "fit_details_{:4.3f}_small_log".format(z_cm) + suffix + '.png'
                logger.info('Saving ' + out_filename)
                plt.savefig(out_filename)

                plt.close()

                plt.plot(self.r_data_cm_1d, dose_mc_MeV_g * self.r_data_cm_1d, 'k.', label="data")
                plt.plot(self.r_data_cm_1d, gauss_data_MeV_g * self.r_data_cm_1d, label="fit")
                plt.legend(loc=0)
                plt.ylabel("dose * r [MeV cm/g]")
                plt.ylim([(dose_mc_MeV_g * self.r_data_cm_1d).min(), (dose_mc_MeV_g * self.r_data_cm_1d).max()])
                plt.yscale('log')
                out_filename = prefix + "fit_details_{:4.3f}_r_log".format(z_cm) + suffix + '.png'
                logger.info('Saving ' + out_filename)
                plt.savefig(out_filename)
                plt.xscale('log')
                out_filename = prefix + "fit_details_{:4.3f}_r_loglog".format(z_cm) + suffix + '.png'
                logger.info('Saving ' + out_filename)
                plt.savefig(out_filename)
                plt.close()

    def _pre_fitting_plots(self, cum_dose_left, z_fitting_cm_1d, dose_fitting_MeV_g_1d, threshold, zmax_cm):
        import matplotlib
        matplotlib.use('Agg')
        import matplotlib.pyplot as plt
        prefix = os.path.join(self.outputdir,
                              '{:s}_plot_{:3.1f}MeV_'.format(
                                  os.path.splitext(os.path.basename(self.ddd_filename))[0],
                                  self.energy_MeV))

        plt.plot(self.z_data_cm_1d, self.dose_data_MeV_g_1d, color='blue', label='dose')
        plt.axvspan(
            0,
            zmax_cm,
            alpha=0.1,
            color='green',
            label="fitting area, covers {:g} % of dose".format(100.0 * (1 - threshold)))
        plt.legend(loc=0)
        plt.xlabel('z [cm]')
        plt.ylabel('dose [a.u.]')
        if self.verbosity > 1:
            out_filename = prefix + 'dose_all.png'
            logger.info('Saving ' + out_filename)
            plt.savefig(out_filename)
        plt.yscale('log')
        out_filename = prefix + 'all_log.png'
        logger.info('Saving ' + out_filename)
        plt.savefig(out_filename)
        plt.close()

        if self.verbosity > 1:
            bp_max_z_pos_cm = self.z_data_cm_1d[self.dose_data_MeV_g_1d == self.dose_data_MeV_g_1d.max()]

            plt.semilogy(self.z_data_cm_1d, cum_dose_left, color='blue', label="cumulative missing dose")
            plt.axvspan(
                0,
                zmax_cm,
                alpha=0.1,
                color='green',
                label="fitting area, covers {:g} % of dose".format(100.0 * (1 - threshold)))
            plt.axhline(threshold, color='black', label='threshold {:g}'.format(threshold))
            plt.axvline(bp_max_z_pos_cm, color='red', label='BP max')
            plt.legend(loc=0)
            plt.xlabel('z [cm]')
            plt.ylabel('fraction of total dose deposited behind z')
            out_filename = prefix + 'dose_frac.png'
            logger.info('Saving ' + out_filename)
            plt.savefig(out_filename)
            plt.close()

        if self.verbosity > 1:
            plt.plot(z_fitting_cm_1d, dose_fitting_MeV_g_1d, 'b', label='dose')
            plt.xlabel('z [cm]')
            plt.ylabel('dose [MeV/g]')
            plt.yscale('log')
            out_filename = prefix + 'dose_log.png'
            logger.info('Saving ' + out_filename)
            plt.savefig(out_filename)
            plt.close()

    def _post_fitting_plots(self, z_fitting_cm_1d,
                            dose_fitting_MeV_g_1d,
                            dz0_MeV_cm_g_data,
                            fwhm1_cm_data,
                            fwhm2_cm_data,
                            weight_data,
                            dz0_MeV_cm_g_error_data,
                            fwhm1_cm_error_data,
                            weight_error_data,
                            fwhm2_cm_error_data):
        import matplotlib.pyplot as plt
        prefix = os.path.join(self.outputdir,
                              '{:s}_plot_{:3.1f}MeV_'.format(
                                  os.path.splitext(os.path.basename(self.ddd_filename))[0],
                                  self.energy_MeV))

        # left Y axis dedicated to FWHM, right one to weight
        fig, ax1 = plt.subplots()
        ax2 = ax1.twinx()
        lns1 = ax1.plot(z_fitting_cm_1d, fwhm1_cm_data, 'g', label='fwhm1')
        upper_fwhm1_line = fwhm1_cm_data + fwhm1_cm_error_data
        lower_fwhm1_line = fwhm1_cm_data - fwhm1_cm_error_data
        ax1.fill_between(z_fitting_cm_1d, lower_fwhm1_line, upper_fwhm1_line,
                         where=upper_fwhm1_line >= lower_fwhm1_line,
                         facecolor='green',
                         alpha=0.1,
                         interpolate=True)
        if self.ngauss == 2:
            lns2 = ax1.plot(z_fitting_cm_1d, fwhm2_cm_data, 'r', label='fwhm2')
            upper_fwhm2_line = fwhm2_cm_data + fwhm2_cm_error_data
            lower_fwhm2_line = fwhm2_cm_data - fwhm2_cm_error_data
            ax1.fill_between(z_fitting_cm_1d, lower_fwhm2_line, upper_fwhm2_line,
                             where=upper_fwhm2_line >= lower_fwhm2_line,
                             facecolor='red',
                             alpha=0.1,
                             interpolate=True)

            lns3 = ax2.plot(z_fitting_cm_1d, weight_data, 'b', label='weight')
            upper_weight_line = weight_data + weight_error_data
            lower_weight_line = weight_data - weight_error_data
            ax2.fill_between(z_fitting_cm_1d, lower_weight_line, upper_weight_line,
                             where=upper_weight_line >= lower_weight_line,
                             facecolor='blue',
                             alpha=0.1,
                             interpolate=True)
            ax2.set_ylabel('weight of FWHM1')
            ax2.set_ylim([0, 1])
        ax1.set_xlabel('z [cm]')
        ax1.set_ylabel('FWHM [cm]')

        # add by hand line plots and labels to legend
        line_objs = lns1
        if self.ngauss == 2:
            line_objs += lns2
            line_objs += lns3
        labels = [l.get_label() for l in line_objs]
        ax1.legend(line_objs, labels, loc=0)

        out_filename = prefix + 'fwhm.png'
        logger.info('Saving ' + out_filename)
        plt.savefig(out_filename)
        plt.close()

        r_step_cm = self.r_data_cm_1d[1] - self.r_data_cm_1d[0]
        r_max_cm = self.r_data_cm_1d[-1] + 0.5 * r_step_cm

        # beam model for single gaussian is following:
        #  G(r, sigma) = 1 / (2pi sigma) * exp( - 0.5 r^2 / sigma^2)
        #  D(z,r) = D(z,0) * G(r, sigma)
        # for double gaussian it is following:
        #  D(z,r) = D(z,0) * ( w * G(r, sigma1) + (1-w) * G(r, sigma2))
        #
        # to get depth dose profile D(z) we need to calculate average dose in some volume at depth z
        # (calculating average dose in a subspace separated by two planes at z=z0 and z=z0+dz will lead to zero dose)
        # we cannot use simple arithmetic mean, as we are dealing with cylindrical scoring and bin mass depends on r
        #
        # let rmax be radius of biggest bin in cylindrical scoring
        # we calculate D(z) which will correspond to depth-dose profile measured with ion. chamber of radius rmax
        # it is basically energy E(z) deposited in slice of radius rmax and thickness dz divided by slice mass
        # D(z) = E(z) / m(z) = E(z) / (pi rmax^2 dz rho)
        # energy E(z) is the sum of energy in all cylindrical shell in a slice and can be calculated as integral
        # thin shell has surface at radius r has surface: 2 pi r dr, thus
        # E(z) = \int_0^rmax D(r,z) rho dz 2 pi r dr
        # finally:
        # D(z) = \int_0^rmax D(r,z) rho dz 2 pi r dr / (pi rmax^2 dz rho)   which leads to:
        #
        # D(z) = 2 / rmax^2 \int_0^rmax D(r,z) r dr
        #
        # for single gaussian model this gives:
        #
        # D(z) = 2 / rmax^2 \int_0^rmax D(z,0) * G(r, sigma) r dr = D(z,0) / rmax^2 \int_0^rmax G(r, sigma) r dr
        #      = D(z,0) / (2 pi sigma rmax^2) \int_0^rmax exp( - 0.5 r^2 / sigma^2) r dr
        #
        # integral \int exp( - 0.5 r^2 / sigma^2) r dr is easy to calculate:
        # https://www.wolframalpha.com/input/?i=%5Cint+exp(+-+0.5+r%5E2+%2F+sigma%5E2)+r+dr
        #
        #  \int exp( - 0.5 r^2 / sigma^2) r dr = -sigma^2 exp( - 0.5 r^2 / sigma^2)
        #
        # which leads to
        #
        # \int_0^rmax exp( - 0.5 r^2 / sigma^2) r dr = sigma^2 ( 1 - exp( - 0.5 rmax^2 / sigma^2))
        #
        # this means depth-dose curve for single gaussian model can be expressed as:
        #
        # D(z) = D(z,0) / (2 pi sigma rmax^2) * sigma^2 ( 1 - exp( - 0.5 rmax^2 / sigma^2))
        #
        # or
        #
        # D(z) = sigma * D(z,0) * ( 1 - exp( - 0.5 rmax^2 / sigma^2)) / (2 pi rmax^2)
        #
        # double gaussian can be calculated in similar way and leads to:
        #
        # D(z) = D(z,0) / (2 pi rmax^2) * ( w * sigma1 * ( 1 - exp( - 0.5 rmax^2 / sigma1^2)) +
        #                                 ( (1-w) * sigma2 * ( 1 - exp( - 0.5 rmax^2 / sigma2^2)))
        #

        if self.ngauss == 1:
            sigma1_cm = fwhm1_cm_data / self._sigma_to_fwhm
            # sigma * D(z,0) / (2 pi rmax^2)
            fit_dose_MeV_g = sigma1_cm * dz0_MeV_cm_g_data / (2.0 * np.pi * r_max_cm ** 2)
            # missing ( 1 - exp( - 0.5 rmax^2 / sigma^2))
            fit_dose_MeV_g *= (np.ones_like(sigma1_cm) - np.exp(-0.5 * r_max_cm / sigma1_cm**2))
            plt.plot(z_fitting_cm_1d, fit_dose_MeV_g, 'r', label='dose fit')
        if self.ngauss == 2:
            sigma1_cm = fwhm1_cm_data / self._sigma_to_fwhm
            sigma2_cm = fwhm2_cm_data / self._sigma_to_fwhm
            w = weight_data

            # ( w * sigma1 * ( 1 - exp( - 0.5 rmax^2 / sigma1^2))
            fit_dose_MeV_g = w * sigma1_cm * \
                (np.ones_like(sigma1_cm) - np.exp(-0.5 * r_max_cm / sigma1_cm**2))

            # ( (1-w) * sigma2 * ( 1 - exp( - 0.5 rmax^2 / sigma2^2)))
            fit_dose_MeV_g += (np.ones_like(w) - w) * sigma2_cm * \
                (np.ones_like(sigma2_cm) - np.exp(-0.5 * r_max_cm / sigma2_cm**2))

            # D(z,0) / (2 pi rmax^2)
            fit_dose_MeV_g *= dz0_MeV_cm_g_data / (2.0 * np.pi * r_max_cm ** 2)
            plt.plot(z_fitting_cm_1d, fit_dose_MeV_g, 'r', label='dose fit')

        plt.plot(z_fitting_cm_1d, dose_fitting_MeV_g_1d, 'b', label='dose MC')
        plt.xlabel('z [cm]')
        plt.ylabel('dose [MeV/g]')
        plt.legend(loc=0)
        out_filename = prefix + 'dose_fit.png'
        logger.info('Saving ' + out_filename)
        plt.savefig(out_filename)
        if self.verbosity > 1:
            plt.yscale('log')
            out_filename = prefix + 'dose_fit_log.png'
            logger.info('Saving ' + out_filename)
            plt.savefig(out_filename)
        plt.close()

    @classmethod
    def gauss_MeV_g(cls, x_cm, amp_MeV_cm_g, sigma_cm):
        return amp_MeV_cm_g / (2.0 * np.pi * sigma_cm) * np.exp(-x_cm ** 2 / (2.0 * sigma_cm ** 2))

    @classmethod
    def gauss_r_MeV_cm_g(cls, x_cm, amp_MeV_cm_g, sigma_cm):
        return cls.gauss_MeV_g(x_cm, amp_MeV_cm_g, sigma_cm) * x_cm

    @classmethod
    def gauss2_MeV_g(cls, x_cm, amp_MeV_cm_g, sigma1_cm, weight, sigma2_add_cm):
        return amp_MeV_cm_g / (2.0 * np.pi) * (
            (weight / sigma1_cm) * np.exp(-x_cm ** 2 / (2.0 * sigma1_cm ** 2))
            + ((1.0 - weight) / (sigma1_cm + sigma2_add_cm))
            * np.exp(-x_cm ** 2 / (2.0 * (sigma1_cm + sigma2_add_cm) ** 2))
        )

    @classmethod
    def gauss2_MeV_g_1st(cls, x_cm, amp_MeV_cm_g, sigma1_cm, weight, sigma2_add_cm):
        return amp_MeV_cm_g / (2.0 * np.pi) * (weight / sigma1_cm) * np.exp(-x_cm ** 2 / (2.0 * sigma1_cm ** 2))

    @classmethod
    def gauss2_MeV_g_2nd(cls, x_cm, amp_MeV_cm_g, sigma1_cm, weight, sigma2_add_cm):
        return amp_MeV_cm_g / (2.0 * np.pi) * ((1.0 - weight) / (sigma1_cm + sigma2_add_cm)) * np.exp(
            -x_cm ** 2 / (2.0 * (sigma1_cm + sigma2_add_cm) ** 2))

    @classmethod
    def gauss2_r_MeV_cm_g(cls, x_cm, amp_MeV_cm_g, sigma1_cm, weight, sigma2_add_cm):
        return cls.gauss2_MeV_g(x_cm, amp_MeV_cm_g, sigma1_cm, weight, sigma2_add_cm) * x_cm

    @classmethod
    def _lateral_fit(cls, r_cm, dose_MeV_g, ngauss=2):
        variance = np.average(r_cm ** 2, weights=dose_MeV_g)

        start_amp_MeV_g = dose_MeV_g.max()
        start_sigma_cm = np.sqrt(variance)

        min_amp_MeV_g = 1e-10 * dose_MeV_g.max()
        min_sigma_cm = 1e-2 * start_sigma_cm

        max_amp_MeV_g = 2.0 * dose_MeV_g.max()
        max_sigma_cm = 1e4 * start_sigma_cm

        from scipy.optimize import curve_fit

        if ngauss == 1:
            try:
                popt, pcov = curve_fit(f=cls.gauss_r_MeV_cm_g,
                                       xdata=r_cm,
                                       ydata=dose_MeV_g * r_cm,
                                       p0=[start_amp_MeV_g, start_sigma_cm],
                                       bounds=([[min_amp_MeV_g, min_sigma_cm], [max_amp_MeV_g, max_sigma_cm]]),
                                       sigma=None)
                # TODO return also parameter errors
                perr = np.sqrt(np.diag(pcov))

                dz0_MeV_cm_g, sigma_cm = popt
                dz0_MeV_cm_g_error, sigma_cm_error = perr
            except RuntimeError as e:
                logger.warning(e)
                dz0_MeV_cm_g, sigma_cm = np.nan, np.nan
                dz0_MeV_cm_g_error, sigma_cm_error = np.nan, np.nan
            factor = 0.0
            factor_error = 0.0
            fwhm2_cm = 0.0
            fwhm2_cm_error = 0.0

        elif ngauss == 2:
            start_weigth = 0.99
            start_sigma2_add_cm = 0.1

            min_weigth = 0.55
            min_sigma2_add_cm = 1e-1

            max_weigth = 1.0 - 1e-12
            max_sigma2_add_cm = 20.0

            try:
                popt, pcov = curve_fit(f=cls.gauss2_r_MeV_cm_g,
                                       xdata=r_cm,
                                       ydata=dose_MeV_g * r_cm,
                                       p0=[start_amp_MeV_g, start_sigma_cm, start_weigth, start_sigma2_add_cm],
                                       bounds=([min_amp_MeV_g, min_sigma_cm, min_weigth, min_sigma2_add_cm],
                                               [max_amp_MeV_g, max_sigma_cm, max_weigth, max_sigma2_add_cm]),
                                       sigma=None)
                perr = np.sqrt(np.diag(pcov))
                dz0_MeV_cm_g_error, sigma_cm_error, factor_error, sigma2_add_cm_error = perr
                dz0_MeV_cm_g, sigma_cm, factor, sigma2_add_cm = popt
            except RuntimeError as e:
                logger.warning(e)
                dz0_MeV_cm_g_error, sigma_cm_error, factor_error, sigma2_add_cm_error = np.nan, np.nan, np.nan, np.nan
                dz0_MeV_cm_g, sigma_cm, factor, sigma2_add_cm = np.nan, np.nan, np.nan, np.nan
            # TODO return also parameter errors
            sigma2_cm = sigma_cm + sigma2_add_cm
            sigma2_cm_error = (sigma_cm_error**2 + sigma2_add_cm_error**2)**0.5
            fwhm2_cm = sigma2_cm * cls._sigma_to_fwhm
            fwhm2_cm_error = sigma2_cm_error * cls._sigma_to_fwhm

        fwhm1_cm = sigma_cm * cls._sigma_to_fwhm

        fwhm1_cm_error = sigma_cm_error * cls._sigma_to_fwhm

        params = fwhm1_cm, factor, fwhm2_cm, dz0_MeV_cm_g
        params_error = fwhm1_cm_error, factor_error, fwhm2_cm_error, dz0_MeV_cm_g_error

        return params, params_error
